"""OpenFOAM module of the API for fluid dynamics."""
import pathlib
from typing import Optional

from inductiva import types
from inductiva.simulation import Simulator


class OpenFOAM(Simulator):
    """Class to invoke a generic DualSPHysics simulation on the API."""

    def __init__(self, api_method: str = "fvm.openfoam.run_simulation"):
        super().__init__()
        self.api_method = api_method

    @property
    def api_method_name(self) -> str:
        return self.api_method

    def run(
        self,
        input_dir: types.Path,
        method_name: str,
        output_dir: Optional[types.Path] = None,
        n_cores: int = 1,
        track_logs: bool = False,
        **openfoam_flags: Optional[str],
    ) -> pathlib.Path:
        """Run the simulation.

        Args:
            n_cores: Number of MPI cores to use for the simulation.
            method_name: OpenFOAM method to run. This involves commands
                from pre-processing, to solvers and post-processing.
            openfoam_flags: Flags to pass to the openfoam method_name.
            other arguments: See the documentation of the base class.
        """
        return super().run(input_dir,
                           output_dir=output_dir,
                           track_logs=track_logs,
                           method_name=method_name,
                           n_cores=n_cores,
                           user_flags=openfoam_flags)

    def run_async(
        self,
        input_dir: types.Path,
        method_name: str,
        n_cores: int = 1,
        **openfoam_flags: Optional[str],
    ) -> str:
        """Run the simulation asynchronously.
        
        Args:
            n_cores: Number of MPI cores to use for the simulation.
            method_name: OpenFOAM method to run. This involves commands
                from pre-processing, to solvers and post-processing.
            openfoam_flags: Flags to pass to the openfoam method_name.
            other arguments: See the documentation of the base class.
            """

        return super().run_async(input_dir,
                                 method_name=method_name,
                                 n_cores=n_cores,
                                 user_flags=openfoam_flags)



"""Wind tunnel scenario to run an object over an air flow."""

from functools import singledispatchmethod
import os
import shutil
from typing import Optional, List, Literal
from enum import Enum
from dataclasses import dataclass

from absl import logging

from inductiva.types import Path
from inductiva.scenarios import Scenario
from inductiva.simulation import Simulator
from inductiva.fluids.simulators import OpenFOAM
from inductiva.utils.templates import (TEMPLATES_PATH,
                                       batch_replace_params_in_template)
from inductiva.utils.files import remove_files_with_tag

SCENARIO_TEMPLATE_DIR = os.path.join(TEMPLATES_PATH, "wind_tunnel")
OPENFOAM_TEMPLATE_INPUT_DIR = "openfoam"


@dataclass
class MeshResolution(Enum):
    """Sets particle radius according to resolution."""
    HIGH = [5, 6]
    MEDIUM = [4, 5]
    LOW = [3, 4]


class WindTunnel(Scenario):
    """Physical scenario of a configurable wind tunnel simulation.
    
    In this scenario, an object is inserted in a wind tunnel described by the
    user. The object is then subject to an air flow determined for which the
    direction and magnitude is defined by the user.
    """

    def __init__(self,
                 flow_velocity: List[float] = None,
                 domain: Optional[dict] = None):
        """Initializes the `WindTunnel` conditions.
        
        Args:
            flow_velocity (dict): Velocity of the air flow in m/s.
            domain (dict): List containing the lower and upper boundary of
                the wind tunnel in each (x, y, z) direction. It is the
                natural description with the default OpenFOAM simulator.
        """
        if flow_velocity is None:
            logging.info("Using a default flow velocity: [30, 0, 0].")
            self.flow_velocity = [30, 0, 0]
        elif len(flow_velocity) != 3:
            raise ValueError("`flow_velocity` must have 3 values.")
        else:
            self.flow_velocity = flow_velocity

        if domain is None:
            logging.info("Using a default domain: `{\"x\":"
                         "[-5, 15], \"y\": [-4, 4], \"z\": [0, 8]}`.")
            self.domain = {"x": [-5, 15], "y": [-4, 4], "z": [0, 8]}
        elif isinstance(domain, dict):
            raise ValueError(
                "`domain` must be a dictionary of the type `{\"x\":"
                "[-5, 15], \"y\": [-4, 4], \"z\": [0, 8]}`.")
        else:
            self.domain = domain

    def simulate(self,
                 simulator: Simulator = OpenFOAM(
                     "windtunnel.openfoam.run_simulation"),
                 output_dir: Optional[Path] = None,
                 object_path: Optional[Path] = None,
                 simulation_time: float = 100,
                 output_time_step: float = 50,
                 resolution: Literal["high", "medium", "low"] = "medium",
                 n_cores: int = 1):
        """Simulates the wind tunnel scenario synchronously.
        
        Args:
            object_path: Path to object inserted in the wind tunnel.
            simulator: Simulator to use for the simulation.
            output_dir: Path to the directory where the simulation output
                is downloaded.
            simulation_time: Simulation time, in seconds.
            write_interval: Interval between simulation outputs, in seconds.
            n_cores: Number of cores to use for the simulation.
            """

        if object_path:
            self.object_path = object_path
        else:
            logging.info("WindTunnel is empty. Object path not specified.")

        self.simulation_time = simulation_time
        self.output_time_step = output_time_step
        self.n_cores = n_cores
        self.resolution = MeshResolution[resolution.upper()].value

        output_path = super().simulate(
            simulator,
            output_dir=output_dir,
            n_cores=n_cores,
            method_name="simpleFoam",
        )

        return output_path

    def simulate_async(self,
                       simulator: Simulator = OpenFOAM(
                           "windtunnel.openfoam.run_simulation"),
                       object_path: Optional[Path] = None,
                       simulation_time: float = 100,
                       output_time_step: float = 50,
                       resolution: Literal["high", "medium", "low"] = "medium",
                       n_cores: int = 1):
        """Simulates the wind tunnel scenario asynchronously.
        
        Args:
            object_path: Path to object inserted in the wind tunnel.
            simulator: Simulator to use for the simulation.
            output_dir: Path to the directory where the simulation output
                is downloaded.
            simulation_time: Simulation time, in seconds.
            write_interval: Interval between simulation outputs, in seconds.
            n_cores: Number of cores to use for the simulation.
            """

        if object_path:
            self.object_path = object_path
        else:
            logging.info("WindTunnel is empty. Object path not specified.")

        self.resolution = MeshResolution[resolution.upper()].value
        self.simulation_time = simulation_time
        self.output_time_step = output_time_step
        self.n_cores = n_cores

        task_id = super().simulate_async(
            simulator,
            n_cores=n_cores,
            method_name="simpleFoam",
        )

        return task_id

    @singledispatchmethod
    @classmethod
    def get_config_filename(cls, simulator: Simulator):  # pylint: disable=unused-argument
        raise ValueError(
            f"Simulator not supported for `{cls.__name__}` scenario.")

    @singledispatchmethod
    def gen_aux_files(self, simulator: Simulator, input_dir: str):
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )

    @singledispatchmethod
    def gen_config(self, simulator: Simulator, input_dir: str):
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )


@WindTunnel.get_config_filename.register
def _(self, simulator: OpenFOAM):  # pylint: disable=unused-argument
    pass


@WindTunnel.gen_aux_files.register
def _(self, simulator: OpenFOAM, input_dir):  # pylint: disable=unused-argument
    # The WindTunnel with OpenFOAM requires changing multiple files
    template_dir = os.path.join(SCENARIO_TEMPLATE_DIR,
                                OPENFOAM_TEMPLATE_INPUT_DIR)

    # Copy all files from the template dir to the input directory
    for directory in os.listdir(template_dir):
        shutil.copytree(os.path.join(template_dir, directory),
                        os.path.join(input_dir, directory))

    # Remove all files that have .jinja in input_dir
    remove_files_with_tag(input_dir, ".jinja")


@WindTunnel.gen_config.register
def _(self, simulator: OpenFOAM, input_dir: str):  # pylint: disable=unused-argument
    """Generates the configuration files for OpenFOAM."""

    # The WindTunnel with OpenFOAM requires changing multiple files
    template_dir = os.path.join(SCENARIO_TEMPLATE_DIR,
                                OPENFOAM_TEMPLATE_INPUT_DIR)

    # Add object path to its respective place
    object_input_dir_path = os.path.join(input_dir, "constant", "triSurface")
    os.mkdir(object_input_dir_path)
    shutil.copy(self.object_path,
                os.path.join(object_input_dir_path, "object.obj"))

    batch_replace_params_in_template(
        templates_dir=template_dir,
        template_filename_paths=[
            os.path.join("0", "include",
                         "initialConditions_template.openfoam.jinja"),
            os.path.join("system", "controlDict_template.openfoam.jinja"),
            os.path.join("system", "blockMeshDict_template.openfoam.jinja"),
            os.path.join("system", "decomposeParDict_template.openfoam.jinja"),
            os.path.join("system", "snappyHexMeshDict_template.openfoam.jinja")
        ],
        params={
            "flow_velocity": self.flow_velocity,
            "simulation_time": self.simulation_time,
            "output_time_step": self.output_time_step,
            "n_cores": self.n_cores,
            "domain": self.domain,
            "resolution": self.resolution,
        },
        output_filename_paths=[
            os.path.join(input_dir, "0", "include", "initialConditions"),
            os.path.join(input_dir, "system", "controlDict"),
            os.path.join(input_dir, "system", "blockMeshDict"),
            os.path.join(input_dir, "system", "decomposeParDict"),
            os.path.join(input_dir, "system", "snappyHexMeshDict")
        ])



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class GCloudExecuter(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.

    Google Cloud executer.
    """

    class MetaOapg:
        required = {
            "vm_type",
            "memory",
            "cpu_count_logical",
            "cpu_count_physical",
            "uuid",
            "host_type",
            "vm_name",
        }

        class properties:
            uuid = schemas.StrSchema
            cpu_count_logical = schemas.IntSchema
            cpu_count_physical = schemas.IntSchema
            memory = schemas.IntSchema

            class host_type(schemas.EnumBase, schemas.StrSchema):

                class MetaOapg:
                    enum_value_to_name = {
                        "gcloud": "GCLOUD",
                    }

                @schemas.classproperty
                def GCLOUD(cls):
                    return cls("gcloud")

            vm_type = schemas.StrSchema
            vm_name = schemas.StrSchema
            __annotations__ = {
                "uuid": uuid,
                "cpu_count_logical": cpu_count_logical,
                "cpu_count_physical": cpu_count_physical,
                "memory": memory,
                "host_type": host_type,
                "vm_type": vm_type,
                "vm_name": vm_name,
            }

    vm_type: MetaOapg.properties.vm_type
    memory: MetaOapg.properties.memory
    cpu_count_logical: MetaOapg.properties.cpu_count_logical
    cpu_count_physical: MetaOapg.properties.cpu_count_physical
    uuid: MetaOapg.properties.uuid
    host_type: MetaOapg.properties.host_type
    vm_name: MetaOapg.properties.vm_name

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["uuid"]
    ) -> MetaOapg.properties.uuid:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["cpu_count_logical"]
    ) -> MetaOapg.properties.cpu_count_logical:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["cpu_count_physical"]
    ) -> MetaOapg.properties.cpu_count_physical:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["memory"]
    ) -> MetaOapg.properties.memory:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["host_type"]
    ) -> MetaOapg.properties.host_type:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["vm_type"]
    ) -> MetaOapg.properties.vm_type:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["vm_name"]
    ) -> MetaOapg.properties.vm_name:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "uuid",
        "cpu_count_logical",
        "cpu_count_physical",
        "memory",
        "host_type",
        "vm_type",
        "vm_name",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["uuid"]
    ) -> MetaOapg.properties.uuid:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["cpu_count_logical"]
    ) -> MetaOapg.properties.cpu_count_logical:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["cpu_count_physical"]
    ) -> MetaOapg.properties.cpu_count_physical:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["memory"]
    ) -> MetaOapg.properties.memory:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["host_type"]
    ) -> MetaOapg.properties.host_type:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["vm_type"]
    ) -> MetaOapg.properties.vm_type:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["vm_name"]
    ) -> MetaOapg.properties.vm_name:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "uuid",
        "cpu_count_logical",
        "cpu_count_physical",
        "memory",
        "host_type",
        "vm_type",
        "vm_name",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        vm_type: typing.Union[
            MetaOapg.properties.vm_type,
            str,
        ],
        memory: typing.Union[
            MetaOapg.properties.memory,
            decimal.Decimal,
            int,
        ],
        cpu_count_logical: typing.Union[
            MetaOapg.properties.cpu_count_logical,
            decimal.Decimal,
            int,
        ],
        cpu_count_physical: typing.Union[
            MetaOapg.properties.cpu_count_physical,
            decimal.Decimal,
            int,
        ],
        uuid: typing.Union[
            MetaOapg.properties.uuid,
            str,
        ],
        host_type: typing.Union[
            MetaOapg.properties.host_type,
            str,
        ],
        vm_name: typing.Union[
            MetaOapg.properties.vm_name,
            str,
        ],
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'GCloudExecuter':
        return super().__new__(
            cls,
            *_args,
            vm_type=vm_type,
            memory=memory,
            cpu_count_logical=cpu_count_logical,
            cpu_count_physical=cpu_count_physical,
            uuid=uuid,
            host_type=host_type,
            vm_name=vm_name,
            _configuration=_configuration,
            **kwargs,
        )



"""Base class for scenarios."""

from abc import ABC, abstractmethod
import tempfile
from typing import Optional
from inductiva.types import Path
from inductiva.simulation import Simulator
from inductiva.utils.files import resolve_path, get_timestamped_path
from inductiva.utils.misc import split_camel_case


class Scenario(ABC):
    """Base class for scenarios."""

    def _setup_output_dir(self, output_dir: str):
        """Setup the scenario output directory."""
        if output_dir is None:
            scenario_name_splitted = split_camel_case(self.__class__.__name__)
            output_dir_prefix = "-".join(scenario_name_splitted).lower()
            output_dir = get_timestamped_path(f"{output_dir_prefix}-output")
        output_dir = resolve_path(output_dir)
        return output_dir

    def _setup_config(self, simulator: Simulator, input_dir: Path):
        """Setup the scenario configuration files and arguments."""
        self.gen_aux_files(simulator, input_dir)
        self.gen_config(simulator, input_dir)

        args = ()
        config_filename = self.get_config_filename(simulator)
        if config_filename:
            args += (config_filename,)
        return args

    @abstractmethod
    def gen_aux_files(self, simulator: Simulator, input_dir: Path):
        """To be implemented in subclasses."""
        pass

    @abstractmethod
    def gen_config(self, simulator: Simulator, input_dir: Path):
        """To be implemented in subclasses."""
        pass

    @abstractmethod
    def get_config_filename(self, simulator: Simulator):
        """To be implemented in subclasses."""
        pass

    def simulate(
        self,
        simulator: Simulator,
        output_dir: Optional[Path] = None,
        **kwargs,
    ) -> Path:
        """Simulates the scenario for a single simulator call."""
        output_dir = self._setup_output_dir(output_dir)
        with tempfile.TemporaryDirectory() as input_dir:
            args = self._setup_config(simulator, input_dir)
            return simulator.run(
                input_dir,
                *args,
                output_dir=output_dir,
                **kwargs,
            )

    def simulate_async(
        self,
        simulator: Simulator,
        **kwargs,
    ) -> str:
        """Simulates the scenario asychronously."""

        with tempfile.TemporaryDirectory() as input_dir:
            args = self._setup_config(simulator, input_dir)
            task_id = simulator.run_async(
                input_dir,
                *args,
                **kwargs,
            )
        return task_id



"""Utilities for parsing script flags."""

from typing import List, Union


def cast_list_to_int(
        list_of_strings: Union[List[str], None]) -> Union[List[int], None]:
    """Converts a list of strings into a list of integers.

    Args:
        list_of_strings: List containing string variables.

    Returns:
        List of integers or None, if the list of strings is None.
    """
    if list_of_strings is None:
        return list_of_strings

    list_of_ints = [int(string) for string in list_of_strings]
    return list_of_ints


def cast_list_to_float(
        list_of_strings: Union[List[str], None]) -> Union[List[float], None]:
    """Converts a list of strings into a list of floats.

    Args:
        list_of_strings: List containing string variables.

    Returns:
        List of floats or None, if the list of strings is None.
    """
    if list_of_strings is None:
        return list_of_strings

    list_of_floats = [float(string) for string in list_of_strings]
    return list_of_floats


def cast_list_to_bool(
        list_of_strings: Union[List[str], None]) -> Union[List[bool], None]:
    """Converts a list of strings into a list of booleans.

    Args:
        list_of_strings: List containing string variables.

    Returns:
        List of booleans or None, if the list of strings is None.
    """
    if list_of_strings is None:
        return list_of_strings

    for string in list_of_strings:
        if string not in ["True", "False", True, False]:
            raise ValueError(f"Could not cast string `{string}` to boolean.")

    list_of_bools = [(str(string) == "True") or (str(string) != "False")
                     for string in list_of_strings]
    return list_of_bools



"""Utils related to template files."""

from typing import Dict, List

from jinja2 import Environment, FileSystemLoader

from inductiva.utils.files import find_path_to_package

TEMPLATES_PATH = find_path_to_package("templates")


def replace_params_in_template(
    templates_dir: str,
    template_filename: str,
    params: Dict,
    output_file_path: str,
) -> None:
    """Replaces parameters in a template file."""

    environment = Environment(loader=FileSystemLoader(templates_dir))
    template = environment.get_template(template_filename)
    stream = template.stream(**params)
    stream.dump(output_file_path)


def batch_replace_params_in_template(
    templates_dir: str,
    template_filename_paths: List[str],
    params: Dict,
    output_filename_paths: List[str],
) -> None:
    """Replaces parameters in a set of template files.
    
    For some simulators, more than one file needs to be changed.
    Moreover, some parameters are altered in different input files.
    To simplify, we can do a batch change for the same params dict.

    Args:
        templates_dir: Directory with the template files.
        template_filename_paths: List containing all template files to
            be changed.
        params: Dictionary of params that are inserted in the template
            files.
        output_filename_paths: List containing the output files in regard of
            the template files.
    """

    for index, template_filename in enumerate(template_filename_paths):
        replace_params_in_template(templates_dir, template_filename, params,
                                   output_filename_paths[index])



"""Post process SPH simulation outputs."""
import os

from base64 import b64encode
from IPython.display import HTML

from inductiva.utils.visualization import create_movie_from_vtk
from inductiva.types import Path


class SPHSimulationOutput:
    """Post process SPH simulation outputs."""

    def __init__(self, sim_output_path: Path):
        """Initializes a `SimulationOutput` object.

        Args:
            sim_output_path: Path to simulation output files.
            """
        self.sim_output_dir = sim_output_path

    def render(self,
               virtual_display: bool = True,
               fps: int = 10,
               color: str = "blue"):
        """Render the simulation as a movie.

        Args:
        fps: Number of frames per second to use in the movie. Renders a
            subset of the vtk files to create the movie.
            Default: 10.
        color: The color of the markers in the simulation.
          If None, the default color is used.
        """

        vtk_dir = os.path.join(self.sim_output_dir, "vtk")

        movie_path = os.path.join(self.sim_output_dir, "movie.mp4")

        create_movie_from_vtk(vtk_dir,
                              movie_path,
                              virtual_display=virtual_display,
                              camera=[(3., 3., 2.), (0., 0., 0.), (1., 1., 2.)],
                              fps=fps,
                              color=color)

        with open(movie_path, "rb") as file_path:
            mp4 = file_path.read()
        movie_url = "data:video/mp4;base64," + b64encode(mp4).decode()

        return HTML(f"""
            <video alt="test" controls>
                <source src="{movie_url}" type="video/mp4">
            </video>
        """)



"""Utilities for creating mesh files for the fluid tank scenario."""

from . import gmsh_utils

from inductiva.fluids.shapes import Cube
from inductiva.fluids.shapes import Cylinder


def create_tank_mesh_file(shape, outlet, path: str):
    """Creates a mesh file for the tank.

    The tank is composed of two blocks:
    - a main (cylindrical/cubic) block representing the tank itself;
    - an optional smaller (cylindrical/cubic) block representing a fluid outlet.
      When present, the top base of this block connects with the bottom base of
      the tank, such that fluid flows freely from the tank to the outlet. The
      bottom base of the outlet is also open, such that flow exits the outlet.

    Both blocks are assumed to have their main axes aligned with the z
    axis.

    Args:
        shape: Shape of the tank.
        outlet: Shape of the outlet. If `None`, no outlet is present.
        path: Path of the file to be created.
    """

    with gmsh_utils.gmshAPIWrapper():
        tank_base_hole_loops = []

        if outlet is not None:

            # Add a circle arc/rectangle loop representing the top base of the
            # outlet. An arc/loop is used instead of a circle/rectangle because
            # this face is not filled, i.e. it is not a surface.
            if isinstance(outlet.shape, Cylinder):
                p_top_outlet, c_top_outlet, l_top_outlet = \
                    gmsh_utils.add_circle_arc(
                        x=outlet.shape.position[0],
                        y=outlet.shape.position[1],
                        z=outlet.shape.position[2] + outlet.shape.height,
                        r=outlet.shape.radius,
                    )
            elif isinstance(outlet.shape, Cube):
                p_top_outlet, c_top_outlet, l_top_outlet = \
                    gmsh_utils.add_z_rectangle_loop(
                        x=outlet.shape.position[0],
                        y=outlet.shape.position[1],
                        z=outlet.shape.position[2] + outlet.shape.dimensions[2],
                        lx=outlet.shape.dimensions[0],
                        ly=outlet.shape.dimensions[1],
                    )

            # Add a circle arc/rectangle loop representing the bottom base of
            # the outlet.
            if isinstance(outlet.shape, Cylinder):
                p_bottom_outlet, c_bottom_outlet, _ = gmsh_utils.add_circle_arc(
                    x=outlet.shape.position[0],
                    y=outlet.shape.position[1],
                    z=outlet.shape.position[2],
                    r=outlet.shape.radius,
                )
            elif isinstance(outlet.shape, Cube):
                p_bottom_outlet, c_bottom_outlet, _ = \
                    gmsh_utils.add_z_rectangle_loop(
                        x=outlet.shape.position[0],
                        y=outlet.shape.position[1],
                        z=outlet.shape.position[2],
                        lx=outlet.shape.dimensions[0],
                        ly=outlet.shape.dimensions[1],
                    )

            # Add the walls of the outlet (cylindrical/cubic) block.
            gmsh_utils.add_cylinder_walls(p_bottom_outlet, c_bottom_outlet,
                                          p_top_outlet, c_top_outlet)

            # Add the loop representing the top base of the outlet to the list
            # of loops representing holes in the bottom base of the tank
            # cylinder.
            tank_base_hole_loops.append(l_top_outlet)

        # Add the top and bottom bases of the tank block, setting the loop
        # representing the top base of the outlet as a hole.
        if isinstance(shape, Cylinder):
            p_top, c_top, _, _ = gmsh_utils.add_circle(
                x=shape.position[0],
                y=shape.position[1],
                z=shape.position[2] + shape.height,
                r=shape.radius,
                hole_loops=[],
            )
            p_bottom, c_bottom, _, _ = gmsh_utils.add_circle(
                x=shape.position[0],
                y=shape.position[1],
                z=shape.position[2],
                r=shape.radius,
                hole_loops=tank_base_hole_loops,
            )

        elif isinstance(shape, Cube):
            p_top, c_top, _, _ = gmsh_utils.add_z_rectangle(
                x=shape.position[0],
                y=shape.position[1],
                z=shape.position[2] + shape.dimensions[2],
                lx=shape.dimensions[0],
                ly=shape.dimensions[1],
                hole_loops=[],
            )
            p_bottom, c_bottom, _, _ = gmsh_utils.add_z_rectangle(
                x=shape.position[0],
                y=shape.position[1],
                z=shape.position[2],
                lx=shape.dimensions[0],
                ly=shape.dimensions[1],
                hole_loops=tank_base_hole_loops,
            )

        # Add the walls of the tank (cylindrical/cubic) block.
        gmsh_utils.add_cylinder_walls(p_bottom, c_bottom, p_top, c_top)

    # Convert the msh file generated by gmsh to obj format.
    gmsh_utils.convert_msh_to_obj_file(path)


def create_tank_fluid_mesh_file(shape, fluid_level, margin, path: str):
    """Creates a mesh file for the fluid.

    The fluid is represented by a block with the same shape as the tank, but
    with a smaller height.

    Args:
        shape: Shape of the tank.
        fluid_level: Height of the fluid.
        margin: Margin to be added to the fluid block.
        path: Path of the file to be created.
    """

    if isinstance(shape, Cube):
        with gmsh_utils.gmshAPIWrapper():
            gmsh_utils.add_box(
                shape.position[0] + margin,
                shape.position[1] + margin,
                shape.position[2] + margin,
                shape.dimensions[0] - 2 * margin,
                shape.dimensions[1] - 2 * margin,
                fluid_level - 2 * margin,
            )

    elif isinstance(shape, Cylinder):
        with gmsh_utils.gmshAPIWrapper():
            gmsh_utils.add_cylinder(
                shape.position[0],
                shape.position[1],
                shape.position[2] + margin,
                shape.radius - margin,
                fluid_level - margin,
            )
    else:
        raise ValueError(f"Invalid fluid shape `{shape}`.")

    # Convert the msh file generated by gmsh to obj format.
    gmsh_utils.convert_msh_to_obj_file(path)



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class InstanceCreate(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "machine_type",
            "image_name",
            "spot",
            "name",
        }

        class properties:
            name = schemas.StrSchema
            machine_type = schemas.StrSchema
            image_name = schemas.StrSchema
            spot = schemas.BoolSchema
            zone = schemas.StrSchema
            __annotations__ = {
                "name": name,
                "machine_type": machine_type,
                "image_name": image_name,
                "spot": spot,
                "zone": zone,
            }

    machine_type: MetaOapg.properties.machine_type
    image_name: MetaOapg.properties.image_name
    spot: MetaOapg.properties.spot
    name: MetaOapg.properties.name

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["name"]
    ) -> MetaOapg.properties.name:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["machine_type"]
    ) -> MetaOapg.properties.machine_type:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["image_name"]
    ) -> MetaOapg.properties.image_name:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["spot"]
    ) -> MetaOapg.properties.spot:
        ...

    @typing.overload
    def __getitem__(
            self, name: typing_extensions.Literal["zone"]
    ) -> MetaOapg.properties.zone:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "name",
        "machine_type",
        "image_name",
        "spot",
        "zone",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["name"]
    ) -> MetaOapg.properties.name:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["machine_type"]
    ) -> MetaOapg.properties.machine_type:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["image_name"]
    ) -> MetaOapg.properties.image_name:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: typing_extensions.Literal["spot"]
    ) -> MetaOapg.properties.spot:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["zone"]
    ) -> typing.Union[MetaOapg.properties.zone, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "name",
        "machine_type",
        "image_name",
        "spot",
        "zone",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        machine_type: typing.Union[
            MetaOapg.properties.machine_type,
            str,
        ],
        image_name: typing.Union[
            MetaOapg.properties.image_name,
            str,
        ],
        spot: typing.Union[
            MetaOapg.properties.spot,
            bool,
        ],
        name: typing.Union[
            MetaOapg.properties.name,
            str,
        ],
        zone: typing.Union[MetaOapg.properties.zone, str,
                           schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'InstanceCreate':
        return super().__new__(
            cls,
            *_args,
            machine_type=machine_type,
            image_name=image_name,
            spot=spot,
            name=name,
            zone=zone,
            _configuration=_configuration,
            **kwargs,
        )



"""Describes the physical scenarios and runs its simulation via API."""

from functools import singledispatchmethod
import os
from typing import List, Literal, Optional
import shutil

from inductiva.types import Path
from inductiva.scenarios import Scenario
from inductiva.simulation import Simulator
from inductiva.fluids.fluid_types import FluidType
from inductiva.fluids.simulators import SPlisHSPlasH
from inductiva.fluids.simulators import DualSPHysics
from inductiva.utils.templates import replace_params_in_template
from inductiva.fluids.scenarios._post_processing import SPHSimulationOutput

TANK_DIMENSIONS = [1, 1, 1]

SPLISHSPLASH_TEMPLATE_FILENAME = "fluid_block_template.splishsplash.json.jinja"
SPLISHSPLASH_CONFIG_FILENAME = "fluid_block.json"
UNIT_BOX_MESH_FILENAME = "unit_box.obj"

DUALSPHYSICS_TEMPLATE_FILENAME = "fluid_block_template.dualsphysics.xml.jinja"
DUALSPHYSICS_CONFIG_FILENAME = "dam_break.xml"


class FluidBlock(Scenario):
    """Physical scenario of a general fluid block simulation."""

    def __init__(self,
                 density: float,
                 kinematic_viscosity: float,
                 dimensions: List[float],
                 position: Optional[List[float]] = None,
                 inital_velocity: Optional[List[float]] = None):
        """Initializes a `FluidBlock` object.

        Args:
            density: Density of the fluid in kg/m^3.
            kinematic_viscosity: Kinematic viscosity of the fluid,
                in m^2/s.
            dimensions: A list containing fluid column dimensions,
                in meters.
            position: Position of the fluid column in the tank,
                in meters.
            initial_velocity: Initial velocity of the fluid block
                in the [x, y, z] axes, in m/s.
        """

        self.fluid = FluidType(density=density,
                               kinematic_viscosity=kinematic_viscosity)

        if len(dimensions) != 3:
            raise ValueError("`fluid_dimensions` must have 3 values.")
        self.dimensions = dimensions

        if position is None:
            self.position = [0, 0, 0]
        else:
            self.position = position

        if inital_velocity is None:
            self.initial_velocity = [0, 0, 0]
        else:
            self.initial_velocity = inital_velocity

    def simulate(
        self,
        simulator: Simulator = DualSPHysics(),
        output_dir: Optional[Path] = None,
        device: Literal["cpu", "gpu"] = "gpu",
        particle_radius: float = 0.02,
        simulation_time: float = 1,
        adaptive_time_step: bool = True,
        particle_sorting: bool = True,
        time_step: float = 0.001,
        output_time_step: float = 1 / 60,
    ):
        """Simulates the scenario.

        Args:
            simulator: Simulator to use.
            output_dir: Directory to store the simulation output.
            device: Device in which to run the simulation.
            particle_radius: Radius of the fluid particles, in meters.
            simulation_time: Simulation time, in seconds.
            adaptive_time_step: Whether to use adaptive time step.
            particle_sorting: Whether to use particle sorting.
            time_step: Time step, in seconds.
            output_time_step: Time step between outputs, in seconds.
        """

        # TODO: Avoid storing these as class attributes.
        self.particle_radius = particle_radius
        self.simulation_time = simulation_time
        self.adaptive_time_step = adaptive_time_step
        self.particle_sorting = particle_sorting
        self.time_step = time_step
        self.output_time_step = output_time_step

        output_path = super().simulate(
            simulator,
            output_dir=output_dir,
            device=device,
        )

        # TODO: Add any kind of post-processing here, e.g. convert files?
        # convert_vtk_data_dir_to_netcdf(
        #     data_dir=os.path.join(output_path, "vtk"),
        #     output_time_step=SPLISHSPLASH_OUTPUT_TIM_STEP,
        #     netcdf_data_dir=os.path.join(output_path, "netcdf"))

        return SPHSimulationOutput(output_path)

    @singledispatchmethod
    @classmethod
    def get_config_filename(cls, simulator: Simulator):  # pylint: disable=unused-argument
        raise ValueError(
            f"Simulator not supported for `{cls.__name__}` scenario.")

    @singledispatchmethod
    def gen_aux_files(self, simulator: Simulator, input_dir: str):
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )

    @singledispatchmethod
    def gen_config(self, simulator: Simulator, input_dir: str):
        raise ValueError(
            f"Simulator not supported for `{self.__class__.__name__}` scenario."
        )


@FluidBlock.get_config_filename.register
def _(cls, simulator: SPlisHSPlasH):  # pylint: disable=unused-argument
    """Returns the configuration filename for SPlisHSPlasH."""
    return SPLISHSPLASH_CONFIG_FILENAME


@FluidBlock.gen_aux_files.register
def _(self, simulator: SPlisHSPlasH, input_dir):  # pylint: disable=unused-argument
    """Generates auxiliary files for SPlisHSPlasH."""
    unit_box_file_path = os.path.join(os.path.dirname(__file__),
                                      UNIT_BOX_MESH_FILENAME)
    shutil.copy(unit_box_file_path, input_dir)


@FluidBlock.gen_config.register
def _(self, simulator: SPlisHSPlasH, input_dir: str):  # pylint: disable=unused-argument
    """Generates the configuration file for SPlisHSPlasH."""

    fluid_margin = 2 * self.particle_radius

    replace_params_in_template(
        templates_dir=os.path.dirname(__file__),
        template_filename=SPLISHSPLASH_TEMPLATE_FILENAME,
        params={
            "simulation_time": self.simulation_time,
            "time_step": self.time_step,
            "particle_radius": self.particle_radius,
            "data_export_rate": 1 / self.output_time_step,
            "tank_filename": UNIT_BOX_MESH_FILENAME,
            "tank_dimensions": TANK_DIMENSIONS,
            "fluid_filename": UNIT_BOX_MESH_FILENAME,
            "fluid": self.fluid,
            "fluid_position": [
                position + fluid_margin for position in self.position
            ],
            "fluid_dimensions": [
                dimension - 2 * fluid_margin for dimension in self.dimensions
            ],
            "fluid_velocity": self.initial_velocity,
            "particle_sorting": self.particle_sorting,
            "adaptive_time_step": self.adaptive_time_step,
        },
        output_file_path=os.path.join(input_dir, SPLISHSPLASH_CONFIG_FILENAME),
    )


@FluidBlock.get_config_filename.register
def _(cls, simulator: DualSPHysics):  # pylint: disable=unused-argument
    """Returns the configuration filename for DualSPHysics."""
    return DUALSPHYSICS_CONFIG_FILENAME


@FluidBlock.gen_aux_files.register
def _(self, simulator: DualSPHysics, input_dir):  # pylint: disable=unused-argument
    pass


@FluidBlock.gen_config.register
def _(self, simulator: DualSPHysics, input_dir: str):  # pylint: disable=unused-argument
    """Generates the configuration file for DualSPHysics."""

    replace_params_in_template(
        templates_dir=os.path.dirname(__file__),
        template_filename=DUALSPHYSICS_TEMPLATE_FILENAME,
        params={
            "simulation_time": self.simulation_time,
            "time_step": self.time_step,
            "particle_distance": 2 * self.particle_radius,
            "output_time_step": self.output_time_step,
            "tank_dimensions": TANK_DIMENSIONS,
            "fluid_dimensions": self.dimensions,
            "fluid_position": self.position,
            "fluid": self.fluid,
            "adaptive_time_step": self.adaptive_time_step,
        },
        output_file_path=os.path.join(input_dir, DUALSPHYSICS_CONFIG_FILENAME),
    )



"""Admin functions to query tasks."""
from typing import Dict, List, Optional

import inductiva
from inductiva import api
from inductiva.client import ApiClient, ApiException
from inductiva.client.apis.tags.admin_api import AdminApi
from inductiva.client.model.task_status_code import TaskStatusCode


def get_tasks_info(statuses: Optional[List[str]] = None,
                   usernames: Optional[List[str]] = None,
                   methods: Optional[List[str]] = None,
                   page=1,
                   per_page=10) -> List[Dict]:
    """Get information about tasks all tasks on the API.

    Tasks can be filtered by status, username, and method_name,
    by providing a list of each as function arguments. Results are
    paginated indexed from 1.
    """
    api_config = api.validate_api_key(inductiva.api_key)

    with ApiClient(api_config) as client:
        api_instance = AdminApi(client)

        query_params = {
            "page": page,
            "per_page": per_page,
        }
        if statuses:
            query_params["status"] = [
                TaskStatusCode(status) for status in statuses
            ]

        if usernames:
            query_params["username"] = usernames

        if methods:
            query_params["method"] = methods

        try:
            # Get User Tasks
            api_response = api_instance.get_tasks(query_params=query_params,)

            return [{**task} for task in api_response.body]

        except ApiException as e:
            raise e



# coding: utf-8

# flake8: noqa
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

__version__ = "1.0.0"

# import ApiClient
from inductiva.client.api_client import ApiClient

# import Configuration
from inductiva.client.configuration import Configuration

# import exceptions
from inductiva.client.exceptions import OpenApiException
from inductiva.client.exceptions import ApiAttributeError
from inductiva.client.exceptions import ApiTypeError
from inductiva.client.exceptions import ApiValueError
from inductiva.client.exceptions import ApiKeyError
from inductiva.client.exceptions import ApiException



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from datetime import date, datetime  # noqa: F401
import decimal  # noqa: F401
import functools  # noqa: F401
import io  # noqa: F401
import re  # noqa: F401
import typing  # noqa: F401
import typing_extensions  # noqa: F401
import uuid  # noqa: F401

import frozendict  # noqa: F401

from inductiva.client import schemas  # noqa: F401


class TaskWithUserInfo(schemas.DictSchema):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    class MetaOapg:
        required = {
            "method_name",
            "task_id",
            "user",
            "status",
        }

        class properties:
            task_id = schemas.StrSchema

            @staticmethod
            def status() -> typing.Type['TaskStatusCode']:
                return TaskStatusCode

            method_name = schemas.StrSchema

            @staticmethod
            def user() -> typing.Type['Username']:
                return Username

            class create_time(
                    schemas.DateTimeBase,
                    schemas.ComposedSchema,
            ):

                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'create_time':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class input_submit_time(
                    schemas.DateTimeBase,
                    schemas.ComposedSchema,
            ):

                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'input_submit_time':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class start_time(
                    schemas.DateTimeBase,
                    schemas.ComposedSchema,
            ):

                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'start_time':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class end_time(
                    schemas.DateTimeBase,
                    schemas.ComposedSchema,
            ):

                class MetaOapg:
                    format = 'date-time'
                    any_of_0 = schemas.StrSchema
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            cls.any_of_0,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'end_time':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            class executer(
                    schemas.ComposedSchema,):

                class MetaOapg:
                    any_of_1 = schemas.NoneSchema

                    @classmethod
                    @functools.lru_cache()
                    def any_of(cls):
                        # we need this here to make our import statements work
                        # we must store _composed_schemas in here so the code is only run
                        # when we invoke this method. If we kept this at the class
                        # level we would get an error because the class level
                        # code would be run when this module is imported, and these composed
                        # classes don't exist yet because their module has not finished
                        # loading
                        return [
                            Executer,
                            cls.any_of_1,
                        ]

                def __new__(
                    cls,
                    *_args: typing.Union[
                        dict,
                        frozendict.frozendict,
                        str,
                        date,
                        datetime,
                        uuid.UUID,
                        int,
                        float,
                        decimal.Decimal,
                        bool,
                        None,
                        list,
                        tuple,
                        bytes,
                        io.FileIO,
                        io.BufferedReader,
                    ],
                    _configuration: typing.Optional[
                        schemas.Configuration] = None,
                    **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                                           frozendict.frozendict, str, date,
                                           datetime, uuid.UUID, int, float,
                                           decimal.Decimal, None, list, tuple,
                                           bytes],
                ) -> 'executer':
                    return super().__new__(
                        cls,
                        *_args,
                        _configuration=_configuration,
                        **kwargs,
                    )

            __annotations__ = {
                "task_id": task_id,
                "status": status,
                "method_name": method_name,
                "user": user,
                "create_time": create_time,
                "input_submit_time": input_submit_time,
                "start_time": start_time,
                "end_time": end_time,
                "executer": executer,
            }

    method_name: MetaOapg.properties.method_name
    task_id: MetaOapg.properties.task_id
    user: 'Username'
    status: 'TaskStatusCode'

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["task_id"]
    ) -> MetaOapg.properties.task_id:
        ...

    @typing.overload
    def __getitem__(
            self,
            name: typing_extensions.Literal["status"]) -> 'TaskStatusCode':
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["method_name"]
    ) -> MetaOapg.properties.method_name:
        ...

    @typing.overload
    def __getitem__(self,
                    name: typing_extensions.Literal["user"]) -> 'Username':
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["create_time"]
    ) -> MetaOapg.properties.create_time:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["input_submit_time"]
    ) -> MetaOapg.properties.input_submit_time:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["start_time"]
    ) -> MetaOapg.properties.start_time:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["end_time"]
    ) -> MetaOapg.properties.end_time:
        ...

    @typing.overload
    def __getitem__(
        self, name: typing_extensions.Literal["executer"]
    ) -> MetaOapg.properties.executer:
        ...

    @typing.overload
    def __getitem__(self, name: str) -> schemas.UnsetAnyTypeSchema:
        ...

    def __getitem__(self, name: typing.Union[typing_extensions.Literal[
        "task_id",
        "status",
        "method_name",
        "user",
        "create_time",
        "input_submit_time",
        "start_time",
        "end_time",
        "executer",
    ], str]):
        # dict_instance[name] accessor
        return super().__getitem__(name)

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["task_id"]
    ) -> MetaOapg.properties.task_id:
        ...

    @typing.overload
    def get_item_oapg(
            self,
            name: typing_extensions.Literal["status"]) -> 'TaskStatusCode':
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["method_name"]
    ) -> MetaOapg.properties.method_name:
        ...

    @typing.overload
    def get_item_oapg(self,
                      name: typing_extensions.Literal["user"]) -> 'Username':
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["create_time"]
    ) -> typing.Union[MetaOapg.properties.create_time, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["input_submit_time"]
    ) -> typing.Union[MetaOapg.properties.input_submit_time, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["start_time"]
    ) -> typing.Union[MetaOapg.properties.start_time, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["end_time"]
    ) -> typing.Union[MetaOapg.properties.end_time, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
        self, name: typing_extensions.Literal["executer"]
    ) -> typing.Union[MetaOapg.properties.executer, schemas.Unset]:
        ...

    @typing.overload
    def get_item_oapg(
            self, name: str
    ) -> typing.Union[schemas.UnsetAnyTypeSchema, schemas.Unset]:
        ...

    def get_item_oapg(self, name: typing.Union[typing_extensions.Literal[
        "task_id",
        "status",
        "method_name",
        "user",
        "create_time",
        "input_submit_time",
        "start_time",
        "end_time",
        "executer",
    ], str]):
        return super().get_item_oapg(name)

    def __new__(
        cls,
        *_args: typing.Union[
            dict,
            frozendict.frozendict,
        ],
        method_name: typing.Union[
            MetaOapg.properties.method_name,
            str,
        ],
        task_id: typing.Union[
            MetaOapg.properties.task_id,
            str,
        ],
        user: 'Username',
        status: 'TaskStatusCode',
        create_time: typing.Union[MetaOapg.properties.create_time, dict,
                                  frozendict.frozendict, str, date, datetime,
                                  uuid.UUID, int, float, decimal.Decimal, bool,
                                  None, list, tuple, bytes, io.FileIO,
                                  io.BufferedReader,
                                  schemas.Unset] = schemas.unset,
        input_submit_time: typing.Union[MetaOapg.properties.input_submit_time,
                                        dict, frozendict.frozendict, str, date,
                                        datetime, uuid.UUID, int, float,
                                        decimal.Decimal, bool, None, list,
                                        tuple, bytes, io.FileIO,
                                        io.BufferedReader,
                                        schemas.Unset] = schemas.unset,
        start_time: typing.Union[MetaOapg.properties.start_time, dict,
                                 frozendict.frozendict, str, date, datetime,
                                 uuid.UUID, int, float, decimal.Decimal, bool,
                                 None, list, tuple, bytes, io.FileIO,
                                 io.BufferedReader,
                                 schemas.Unset] = schemas.unset,
        end_time: typing.Union[MetaOapg.properties.end_time, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, bool,
                               None, list, tuple, bytes, io.FileIO,
                               io.BufferedReader,
                               schemas.Unset] = schemas.unset,
        executer: typing.Union[MetaOapg.properties.executer, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, bool,
                               None, list, tuple, bytes, io.FileIO,
                               io.BufferedReader,
                               schemas.Unset] = schemas.unset,
        _configuration: typing.Optional[schemas.Configuration] = None,
        **kwargs: typing.Union[schemas.AnyTypeSchema, dict,
                               frozendict.frozendict, str, date, datetime,
                               uuid.UUID, int, float, decimal.Decimal, None,
                               list, tuple, bytes],
    ) -> 'TaskWithUserInfo':
        return super().__new__(
            cls,
            *_args,
            method_name=method_name,
            task_id=task_id,
            user=user,
            status=status,
            create_time=create_time,
            input_submit_time=input_submit_time,
            start_time=start_time,
            end_time=end_time,
            executer=executer,
            _configuration=_configuration,
            **kwargs,
        )


from inductiva.client.model.executer import Executer
from inductiva.client.model.task_status_code import TaskStatusCode
from inductiva.client.model.username import Username



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from inductiva.client.paths.gcp_instances.post import CreateInstance
from inductiva.client.paths.gcp_instances_group.post import CreateInstanceGroup
from inductiva.client.paths.gcp_instances.delete import DeleteInstance
from inductiva.client.paths.gcp_instances_group.delete import DeleteInstanceGroup
from inductiva.client.paths.gcp_instances_price.get import GetInstancePrice


class InstanceApi(
        CreateInstance,
        CreateInstanceGroup,
        DeleteInstance,
        DeleteInstanceGroup,
        GetInstancePrice,
):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    pass



from inductiva.client.paths.gcp_instances.post import ApiForpost
from inductiva.client.paths.gcp_instances.delete import ApiFordelete


class GcpInstances(
        ApiForpost,
        ApiFordelete,
):
    pass



"""GROMACS module of the API"""

import pathlib
from typing import Optional, List

from inductiva import types
from inductiva.simulation import Simulator


class GROMACS(Simulator):
    """Class to invoke any GROMACS command on the API."""

    @property
    def api_method_name(self) -> str:
        return "md.gromacs.run_simulation"

    def run(
        self,
        input_dir: types.Path,
        commands: List[dict],
        track_logs: bool = False,
        output_dir: Optional[types.Path] = None,
    ) -> pathlib.Path:
        """Run a list of GROMACS commands.

        Args:
            input_dir: Path to the directory containing the input files.
            commands: List of commands to run using the GROMACS simulator.
            track_logs: If True, the logs of the remote execution will be
            logged. 
            output_dir: Path to the directory where the output files will be
            stored. If not provided, a timestamped directory will be created.
        """
        return super().run(input_dir,
                           output_dir=output_dir,
                           track_logs=track_logs,
                           commands=commands)

    def run_async(
        self,
        input_dir: types.Path,
        commands: List[dict],
    ) -> str:
        """Run a list of GROMACS commands asynchronously.

        Args:
            input_dir: Path to the directory containing the input files.
            commands: List of commands to run using the GROMACS simulator.
        """
        return super().run_async(input_dir, commands=commands)



# coding: utf-8
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""

from inductiva.client.paths.tasks_task_id_output.get import DownloadTaskOutput
from inductiva.client.paths.tasks_task_id.get import GetTask
from inductiva.client.paths.tasks_task_id_status.get import GetTaskStatus
from inductiva.client.paths.tasks.get import GetUserTasks
from inductiva.client.paths.tasks_task_id_kill.post import KillTask
from inductiva.client.paths.tasks_submit.post import SubmitTask
from inductiva.client.paths.tasks_task_id_input.post import UploadTaskInput


class TasksApi(
        DownloadTaskOutput,
        GetTask,
        GetTaskStatus,
        GetUserTasks,
        KillTask,
        SubmitTask,
        UploadTaskInput,
):
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """
    pass



"""Client for Inductiva's web API."""
import os

from . import fluids
from . import core
from . import molecules
from . import tasks
from . import admin

from absl import logging

api_url = os.environ.get("INDUCTIVA_API_URL", "http://api.inductiva.ai")
output_dir = os.environ.get("INDUCTIVA_OUTPUT_DIR", "inductiva_output")
api_key = os.environ.get("INDUCTIVA_API_KEY")
working_dir = None

logging.set_verbosity(logging.INFO)



"""Visualization processing of WindTunnel scenario.

This class implements various visualization capabilities for
the WindTunnel scenario. Namely:
    - Pressure over object; 
    - Cutting plane;
    - StreamLines.

Currently, we only support the OpenFOAM simulator.
"""
import os
from dataclasses import dataclass
from enum import Enum
from typing import Literal

import pyvista as pv

from inductiva.types import Path
from inductiva.utils.visualization import MeshData


class WindTunnelSimulationOutput:
    """Post-Process WindTunnel simulation outputs.
    
    Current Support:
        OpenFOAM
    """

    def __init__(self, sim_output_path: Path, time_step: int = 100):
        """Initializes a `WindTunnelSimulationOutput` object.

        Args:
            sim_output_path: Path to simulation output files.
            time_step: Time step where we read the data.
        """

        self.sim_output_path = sim_output_path
        self.time_step = time_step
        self.object_data = self.get_object_data()

    def get_object_data(self):  # pylint: disable=unused-argument
        """Get aerodynamics data over an object inside the WindTunnel.
        
        Current Support - OpenFOAM
        """

        reading_file = os.path.join(self.sim_output_path, "foam.foam")

        # Create reading file
        with open(reading_file, "w", encoding="utf-8") as file:
            file.close()

        # Initialize reader and define reading time-step
        reader = pv.POpenFOAMReader(reading_file)
        reader.set_active_time_value(self.time_step)

        mesh = reader.read()
        object_data = mesh["boundary"]["object"]

        return object_data

    def get_physical_field(self, physical_property: str = "pressure"):
        """Get a physical scalar field over mesh points for a certain time_step.
        
        Returns:
            A MeshData object that allow to manipulate the data over a mesh
            and to render it.
        """
        property_notation = OpenFOAMPhysicalProperty[
            physical_property.upper()].value
        physical_field = MeshData(self.object_data, property_notation)

        return physical_field

    def get_streamlines(self,
                        physical_property: Literal["pressure",
                                                   "velocity"] = "pressure"):
        """Get streamlines over the object in the WindTunnel."""

        streamlines_path = os.path.join(self.sim_output_path, "postProcessing",
                                        "sets", "streamLines",
                                        str(self.time_step))

        property_notation = OpenFOAMPhysicalProperty[
            physical_property.upper()].value
        streamlines_file = "track0_" + property_notation + ".vtk"

        streamlines_mesh = pv.read(
            os.path.join(streamlines_path, streamlines_file))

        return streamlines_mesh

    def get_flow_plane(self,
                       physical_property: Literal["pressure",
                                                  "velocity"] = "pressure"):
        """Get flow properties in a plane of the domain in WindTunnel."""

        cutting_plane_path = os.path.join(self.sim_output_path,
                                          "postProcessing", "cuttingPlane",
                                          str(self.time_step))

        # Obtain notation for the physical property for the simulator.
        property_notation = OpenFOAMPhysicalProperty[
            physical_property.upper()].value
        cutting_plane_file = property_notation + "_yNormal.vtk"

        cutting_plane_mesh = pv.read(
            os.path.join(cutting_plane_path, cutting_plane_file))

        return cutting_plane_mesh

    def render_flow(self,
                    flow_property_mesh,
                    physical_property: Literal["pressure",
                                               "velocity"] = "pressure",
                    background_color: str = "black",
                    flow_cmap: str = "viridis",
                    object_color: str = "white",
                    save_path: Path = None):
        """Render flow property over the object in the WindTunnel."""

        # Obtain notation for the physical property for the simulator.
        property_notation = OpenFOAMPhysicalProperty[
            physical_property.upper()].value

        plotter = pv.Plotter()
        plotter.background_color = background_color
        plotter.add_mesh(self.object_data, color=object_color)
        plotter.add_mesh(flow_property_mesh,
                         scalars=property_notation,
                         cmap=flow_cmap)
        plotter.view_xz()
        plotter.show(screenshot=save_path)
        plotter.close()


@dataclass
class OpenFOAMPhysicalProperty(Enum):
    """Defines the notation used for physical properties in OpenFOAM."""
    PRESSURE = "p"
    VELOCITY = "U"
